from pathlib import Path
import json
import os


class Config:
    _PAPR_VERSION = "0.0.19"

    def __init__(self):
        self.config_path = str(Path.home()) + "/.papr"
        self.config_file = self.config_path + "/papr.cfg"
        self._read_config()
        self._latest_version = ""

    def get(self, key, default=""):
        return self.config.get(key, default)

    def _read_config(self):
        if not os.path.exists(self.config_path):
            os.mkdir(self.config_path)
        if not os.path.exists(self.config_file):
            self._create_default()
        f = open(self.config_file, "r")
        self.config = json.loads(f.read())
        f.close()

    def _create_default(self):
        self.config = {
            "cfg_version": "0.0.2",
            "default_repo": "null",
            "viewer": "/usr/bin/evince",
            "last_version_started": self.papr_version(),
            "browser": "firefox",
            "browser_params": []
        }
        self._write_config()

    def _write_config(self):
        f = open(self.config_file, "w")
        f.write(json.dumps(self.config))
        f.close()

    def get_export_path(self, default):
        return self.get("export_path", default)

    def set_export_path(self, path):
        self.config["export_path"] = path
        self._write_config()

    def get_viewer(self):
        return self.get("viewer", "/usr/bin/evince")

    def set_viewer(self, path_viewer):
        self.config["viewer"] = path_viewer
        self._write_config()

    def set_default_repo(self, path):
        self.config["default_repo"] = path
        self._write_config()

    def latest_version(self):
        return self._latest_version

    def set_latest_version(self, version):
        self._latest_version = version

    def papr_version(self):
        return self._PAPR_VERSION

    def new_features_already_shown(self):
        key = "last_version_started"
        v = self.papr_version()
        last_version_started = self.config.get(key, "")
        if last_version_started != v:
            self.config[key] = v
            self._write_config()
            return False
        return True

    def get_browser(self):
        return self.get("browser", "firefox")

    def get_browser_params(self):
        return self.get("browser_params", [])

    def set_browser(self, browser):
        self.config["browser"] = browser
        self._write_config()

    def set_browser_params(self, params):
        self.config["browser_params"] = params
        self._write_config()



import sys
import pathlib
import yaml

from papr.lib.config import Config
from papr.lib.repository import Repository, DATA_PATH, META_FILE, path_for_paper_data
from .v1_repository import RepositoryV1, REPO_META


def assert_migration_not_needed(conf: Config):
    if _migration_needed(conf):
        print("Migration needed.")
        print("Start 'papr migrate' first.")
        sys.exit(1)


def do_migration(conf: Config):
    if not _migration_needed(conf):
        print("Migration not needed.")
        sys.exit(1)
    print("Migrating database... please wait and do not interrupt this process.")
    _do_migration(RepositoryV1(conf), conf)
    print("DONE. You can now use papr.")


def _migration_needed(conf: Config):
    r = Repository(conf)
    return not r.is_v2_repo()


def _write_file(path, filename, data):
    path.mkdir(parents=True, exist_ok=True)
    p = path.joinpath(filename)
    with open(p, "wt") as f:
        f.write(data)


def _do_migration(repo: RepositoryV1, conf: Config):
    repo_path = pathlib.Path(repo.pdf_path())
    data_path = repo_path.joinpath(DATA_PATH)  # for notes, summaries and metadata
    data_path.mkdir(parents=True, exist_ok=True)

    v2 = Repository(conf)

    for paper_v1 in repo.list():
        # path = <repo_path>/data/p00000_title_of_paper/
        path = path_for_paper_data(paper_v1, data_path)
        if paper_v1.has_summary():
            _write_file(path, "summary.md", paper_v1.summary())
            paper_v2 = v2.get_paper(paper_v1.idx())
            paper_v2.set_has_summary(True)
            v2.update_paper(paper_v2)
        if paper_v1.has_notes():
            _write_file(path, "notes.md", paper_v1.msg())
            paper_v2 = v2.get_paper(paper_v1.idx())
            paper_v2.set_has_notes(True)
            v2.update_paper(paper_v2)

    # store repository version into a file <pdf_path>/.paper/meta.yml
    meta_path = repo_path.joinpath(REPO_META).joinpath(META_FILE)
    data = {
        "repo_version": "2"
    }
    with open(meta_path, "wt") as f:
        yaml.dump(data, f)



import sys
from .termout import _hr


def command():
    s = sys.argv[0]
    return s if s.rfind("/") < 0 else s[s.rfind("/")+1:]


def help(exitcode=0):
    print("Usage: " + command() + " [search regex] | <command> [<args>]")
    print("")
    print("Start papr without any arguments to start the text UI.")
    print("")
    print("Commands")
    _hr()
    print("  help       This help message.")
    print("\nRepository commands")
    print("  init       Create a new repository in the current directory and sets it to")
    print("             the default repository.\n")
    print("  default    Set the current repository as default repository.")
    print("")
    print("  config key value  Change configuration.")
    print("             e.g. config viewer /usr/bin/okular")
    print("")
    print("List papers")
    print("  list       List all tracked papers in the current repository.\n")
    print("  search <regex>")
    print("             List all papers for which the regex matches the title.")
    print("")
    print("Reading papers")
    print("  read <id>  Read paper with given ID.\n")
    print("  last [<n>] Read the paper with the n-th largest ID. default: n = 1")
    print("")
    print("Adding papers")
    print("  fetch  <file> [title]")
    print("             Add a paper to the repository.  The file is copied to the")
    print("             repository. Filename will be <idx>_<normalized_title>.pdf")
    print("             If you don't specify a title you will be asked for it.")
    print("")
    print("         <url> [title]")
    print("             Download a paper and add it to the repository. If the URL points")
    print("             to a pdf title is required. If the URL points to a web page which")
    print("             contains the title and a link to the pdf in a recognized format")
    print("             title is not required and the pdf will be downloaded.")
    print("")
    print("         --urls <filename>")
    print("             Load papers for the URLs stored in a file. (one URL per line)")
    print("")
    print("  add    <file> [idx]")
    print("             Add a file which is located in the repository directory but which")
    print("             is not tracked yet. The filename does not change and will be used")
    print("             as the title without the extension.")
    print("Export")
    print("  html       Export as HTML.")
    print("  export     Export database with summaries and notes as YML.")
    print()
    sys.exit(exitcode)



import os
import sys
from pathlib import Path

import termcolor

from .latest_version import latest_version
from .config import Config
from .console import cursor_on, cursor_off, cursor_top_left, cursor_up, cursor_down, write_xy
from .edit import edit_notes_of_paper, tags_of_paper, abstract_of_paper, details_of_paper, list_of_tags, edit_title, \
    edit_summary_of_paper, show_summaries
from .html import export_repository_html
from .termin import read_key
from .termout import rows, empty_line, print_paper, cols, write
from .tools import filter_list, show_pdf, filter_list_re, highlight_query, show_in_browser
from .cmd_fetch import NEWTAG
from .ui_scrollview import ScrollView
from .repository import Repository


def expand_to_colwidth(s: str):
    c = cols()
    return s + (" " * (c - len(s)))


def hr():
    return "─" * cols()


def build_header():
    return "Id    Title " + " " * (cols() - 12 - 8) + "Stars" + "   " + hr()


def colored(s, fg, bg, bold=False):
    attr = ["bold"] if bold else []
    return termcolor.colored(s, fg, bg, attrs=attr)


def build_default_header():
    return termcolor.colored(
        expand_to_colwidth("ESC/q: quit | ENTER: open | ↑: up | ↓: down | s: search | n: notes | h: help"),
        "white", attrs=["bold"])


def build_search_header():
    return termcolor.colored(
        expand_to_colwidth("ESC: cancel + back to selection mode | ENTER: back to selection mode"),
        "white", "on_yellow", attrs=["bold"])


def build_title(conf):
    current_v = conf.papr_version()
    s = "papr " + current_v
    v = conf.latest_version()
    if len(v) > 0 and v != current_v:
        s += " (new version available)"
    return termcolor.colored(
        expand_to_colwidth(s),
        "white", "on_blue", attrs=["bold"])


def extent(s, width):
    return s + (" " * (width - len(s)))


# TODO copy&paste
def write_box_xy(x, y, content, select_lineno=None, wmax=None, arrow_down=False, arrow_up=False):
    if not wmax:
        wmax = max([len(c) for c in content])
    write_xy(x, y, colored("┌" + ("─" * wmax) + "┐", "white", "on_blue"))
    y += 1
    for idx, s in enumerate(content):
        write_xy(x, y, colored("│", "white", "on_blue"))
        if idx == select_lineno:
            write(colored(extent(s, wmax), "white", "on_red", True))
        else:
            write(colored(extent(s, wmax), "white", "on_blue"))
        if arrow_up and idx == 0:
            write(colored("↑", "blue", "on_white"))
        elif arrow_down and idx == len(content) - 1:
            write(colored("↓", "blue", "on_white"))
        else:
            write(colored("│", "white", "on_blue"))
        y += 1
    write_xy(x, y, colored("└" + ("─" * wmax) + "┘", "white", "on_blue"))


# content: List of strings
def write_box(content, select_lineno=None, wmax=None, arrow_down=False, arrow_up=False):
    if not wmax:
        wmax = max([len(c) for c in content])
    write(colored("\r┌" + ("─" * wmax) + "┐", "white", "on_blue"))
    for idx, s in enumerate(content):
        cursor_down(1)
        write(colored("\r│", "white", "on_blue"))
        if idx == select_lineno:
            write(colored(extent(s, wmax), "white", "on_red", True))
        else:
            write(colored(extent(s, wmax), "white", "on_blue"))
        if arrow_up and idx == 0:
            write(colored("↑", "blue", "on_white"))
        elif arrow_down and idx == len(content) - 1:
            write(colored("↓", "blue", "on_white"))
        else:
            write(colored("│", "white", "on_blue"))
    cursor_down(1)
    write(colored("\r└" + ("─" * wmax) + "┘", "white", "on_blue"))


def redraw(state, papers, v, conf):
    cursor_top_left()
    write(build_title(conf))

    if state.in_search or state.in_re_search:
        write(build_search_header())
    else:
        write(build_default_header())
    write(build_header())

    cnt = 0 + 3
    for idx, p in enumerate(papers[v.first():v.end()]):
        print_paper(p, idx + v.first() == v.selected())
        cnt += 1

    # clear remaining lines
    n_rows = rows()
    while cnt < n_rows - 2:
        cnt += 1
        write(empty_line())

    # Now, the cursor is on the last line. Clear that line and go to the first column.
    sys.stdout.write(empty_line() + "\r")

    if state.in_search or state.in_re_search:
        write("Search: " + state.search + "▃")
    elif state.in_filter:
        if state.select_tag:
            write("Select a tag.")
            wmax = max([len(tag) for tag, _ in state.tags])
            tags = [extent(tag, wmax) for tag, _ in state.tags[state.tag_scrollview.first():state.tag_scrollview.end()]]
            cursor_up(2+state.tag_scrollview.rows())
            write_box(tags, state.tag_scrollview.selected() - state.tag_scrollview.first())
        else:
            write("Filter by: (t)ags")
    elif state.in_help:
        h = [
            " Key bindings",
            " ",
            " q / ESC       : Quit",
            " ENTER         : open the paper in a PDF reader.",
            " a             : Show abstract for paper.",
            " i / Arrow up  : Move selection up.",
            " k / Arrow down: Move selection down.",
            " s             : Live search in title.",
            " r             : Live search in title via regular expression. ",
            " n             : Edit notes.",
            " t             : Edit tags.",
            " e             : Edit title.",
            " 0..5          : Set voting.",
            " y             : Show all stored information about a paper.",
            " l             : Show statistics about tags.",
            " f             : Filter (e.g. by tags).",
            " F             : Clear filter.",
            " c             : Hide/show paper.",
            " .             : Show hidden/visible papers.",
            " space         : Open a context menu.",
            " m             : List all papers with a summary.",
            " b             : Open web browser for details."
        ]
        n = min(len(h), n_rows-2-1)   # number of lines in the box
        offset = state.in_help_offset # index of element in h which should be the first line in the box
        offset = min(offset, len(h)-n)
        state.in_help_offset = offset
        cursor_up(2+n)                # 2 lines for frame + n lines for text
        width = max([len(c) for c in h])
        arr_up = offset > 0
        arr_down = offset < len(h) - n
        write_box(h[offset:offset+n], wmax=width, arrow_down=arr_down, arrow_up=arr_up)
    else:
        if len(state.selected_tag) > 0:
            write("Tag: " + state.selected_tag)
        if len(state.selected_tag) > 0 and len(state.search) > 0:
            write(" | ")
        if len(state.search) > 0:
            write("Search: " + state.search)


class State:
    def __init__(self):
        self.search = ""
        self.in_search = False
        self.in_re_search = False

        self.in_filter = False
        self.select_tag = False
        self.tag_scrollview = None
        self.tags = []
        self.selected_tag = ""

        self.in_help = False
        self.in_help_offset = 0

        self.show_hidden = False
        self.show_menu = False


def center_box(content):
    posy = max(0, rows() // 2 - len(content) // 2)
    posx = max(0, cols() // 2 - max([len(i) for i in content]) // 2)
    return posx, posy


def show_new_features(conf: Config):
    message = [
        "CHANGES FROM 0.0.18 TO 0.0.19",
        "-----------------------------",
        "• Navigation",
        "  • Shift+Up/Down - jump 5 entries up/down",
        "  • Page Up/Down  - jump one page up/down"
        "• Fetch",
        "  • You can now specify more than just one URL for fetch command.",
        "  • You can now specify tags for fetch command.",
        "• Storage for notes and summary more open and accessible. You can now",
        "  read and edit your notes and summaries without papr.",
        "• Export to HTML.",
        "• Minor improvements to UI."
#        "",
#        "CHANGES FROM 0.0.17 TO 0.0.18",
#        "-----------------------------",
#        "• You can now change the title of a paper with the key 'e'.",
#        "• Never miss a new version! If a new version is available, it will be",
#        "  shown at the top of the screen.",
#        "  For this the setup.py from the GitHub repository is retrieved and",
#        "  checked.",
#        "• When you upgrade papr to the latest version this message with changes",
#        "  is shown on the first start.",
#        "• You can hide papers with the key 'c'. Use the same key to make hidden",
#        "  papers visible again. You can change between the list of hidden and",
#        "  visible papers with the key '.'",
#        "• Press 'space' to get a context menu.",
#        "• You can already create notes for a paper. Now, you can also create a",
#        "  summary. Just select 'edit summary' from the context menu or use the",
#        "  shortcut 'U'. List all papers with a summary with the shortcut 'm'."
    ]
    posx, posy = center_box(message)
    write_box_xy(posx, posy, message)


def select_item(menu):
    posx, posy = center_box(menu)
    selection = 0
    while True:
        write_box_xy(posx, posy, menu, select_lineno=selection)
        k = read_key()
        if ord(k) == 27 or k == 'q':
            return None
        elif ord(k) == 10:  # Enter
            return selection
        elif k == 'i' or k == '#':
            selection = max(0, selection - 1)
        elif k == 'k' or k == '+':
            selection = min(len(menu) - 1, selection + 1)


def ui_main_or_search_loop(papers_in, repo: Repository, conf: Config):
    n_papers = len(papers_in)
    n_view_rows = rows() - 3 - 1 - 1
    # -3 = 3 rows for header
    # -1 = 1 row for search line
    # -1 = 1 "papr" header line
    v = ScrollView(n_elements=n_papers, rows=n_view_rows, selected=n_papers-1)

    s = State()
    papers = papers_in[:]

    while True:
        if s.show_hidden:
            filtered_papers = [p for p in papers if p.hidden()]
            if len(filtered_papers) != len(papers):
                papers = filtered_papers
                pos = min(v.selected(), len(papers) - 1)
                v = ScrollView(n_elements=len(papers), rows=n_view_rows, selected=pos)
        else:
            # Filter hidden papers
            filtered_papers = [p for p in papers if not p.hidden()]
            if len(filtered_papers) != len(papers):
                papers = filtered_papers
                pos = min(v.selected(), len(papers) - 1)
                v = ScrollView(n_elements=len(papers), rows=n_view_rows, selected=pos)



        cursor_off()
        redraw(s, papers, v, conf)

        # If a new version is started for the first time show the new features.
        if not conf.new_features_already_shown():
            show_new_features(conf)
            read_key()
            continue

        # Menu
        k = None
        if s.show_menu:
            values = [
                [" Edit notes                  ", "n"],
                [" Edit title                  ", "e"],
                [" Edit tags                   ", "t"],
                [" Edit summary                ", "U"],
                [" Show abstract               ", "a"],
                [" Show details                ", "y"]
            ]
            menu = [i[0] for i in values]
            k = select_item(menu)
            if k is not None:
                k = values[k][1]
            s.show_menu = False
        else:
            k = read_key()

        if k is None or k == '~':
            continue

        if s.in_search or s.in_re_search:
            for p in papers:
                p.set_highlights([])
            f = filter_list_re if s.in_re_search else filter_list

            if ord(k) == 27:
                s.in_search = s.in_re_search = False
                s.search = ""
            elif ord(k) == 10:
                s.in_search = s.in_re_search = False
            elif ord(k) == 127:
                s.search = s.search[:-1]
            else:
                s.search += k

            papers = f(papers_in, s.search)
            # highlight search
            if s.in_search:
                for p in papers:
                    positions = highlight_query(p, s.search)
                    p.set_highlights(positions)

            if len(s.selected_tag) > 0:
                papers = [p for p in papers if s.selected_tag in p.tags()]
            v = ScrollView(n_elements=len(papers), rows=n_view_rows, selected=len(papers)-1)
        elif s.in_filter:
            if s.select_tag:
                if ord(k) == 27:
                    s.in_filter = s.select_tag = False
                elif k == '#':
                    s.tag_scrollview.up()
                elif k == '+':
                    s.tag_scrollview.down()
                elif k == '[':
                    s.tag_scrollview.up(5)
                elif k == ']':
                    s.tag_scrollview.down(5)
                elif k == '§':
                    s.tag_scrollview.pagedown()
                elif k == '$':
                    s.tag_scrollview.pageup()
                elif ord(k) == 10:
                    s.selected_tag = s.tags[s.tag_scrollview.selected()][0]
                    papers = [p for p in papers_in if s.selected_tag in p.tags()]
                    v = ScrollView(n_elements=len(papers), rows=n_view_rows, selected=len(papers) - 1)
                    s.in_filter = False
            else:
                if k == 't':
                    s.select_tag = True
                    s.tags = repo.all_tags()
                    n = len(s.tags)
                    s.tag_scrollview = ScrollView(n_elements=n, rows=min(n_view_rows-2, n), selected=n-1)
                else:
                    s.in_filter = False
        elif s.in_help:
            if k == '#':
                s.in_help_offset = max(0, s.in_help_offset - 1)
            elif k == '+':
                s.in_help_offset += 1
            else:
                s.in_help = False
        else:
            if k == 'l':
                list_of_tags(repo)
            elif k == 'a':
                abstract_of_paper(papers[v.selected()])
            elif k == 'y':
                details_of_paper(papers[v.selected()])
            elif k == 'i' or k == '#':
                v.up()
            elif k == 'k' or k == '+':
                v.down()
            elif k == '[':
                v.up(5)
            elif k == ']':
                v.down(5)
            elif k == '§':
                v.pagedown()
            elif k == '$':
                v.pageup()
            elif k == 'n':
                edit_notes_of_paper(repo, papers[v.selected()])
            elif k == 'U':
                edit_summary_of_paper(repo, papers[v.selected()])
            elif k == 't':
                tags_of_paper(repo, papers[v.selected()])
            elif ord(k) == 10:
                p = v.selected()
                papers[p].remove_tag(NEWTAG)
                repo.update_paper(papers[p])
                show_pdf(papers[p], repo.pdf_path(), conf.get_viewer())
            elif ord(k) == 27 or k == 'q':
                break
            elif k == 's':
                s.in_search = True
                # copy&paste from above
                for p in papers:
                    positions = highlight_query(p, s.search)
                    p.set_highlights(positions)
            elif k == 'r':
                s.in_re_search = True
            elif '0' <= k <= '5':
                stars = int(k)
                papers[v.selected()].set_stars(stars)
                repo.update_paper(papers[v.selected()])
            elif k == 'f':
                s.in_filter = True
                s.select_tag = False
            elif k == 'F':
                s.selected_tag = ""
                papers = papers_in[:]
                v = ScrollView(n_elements=len(papers), rows=n_view_rows, selected=len(papers) - 1)
                # TODO: consider search
            elif k == 'h':
                s.in_help = True
            elif k == 'e':
                edit_title(papers[v.selected()], repo)
            elif k == 'c':
                p = papers[v.selected()]
                p.hide(not p.hidden())
                repo.update_paper(p)
            elif k == '.':
                s.show_hidden = not s.show_hidden
                papers = papers_in[:]
            elif k == ' ':
                s.show_menu = True
            elif k == 'm':
                show_summaries(repo, cols())
            elif k == 'x':
                file_dialog(repo, conf)
            elif k == 'b':
                show_in_browser(repo, papers[v.selected()], conf.get_browser(), conf.get_browser_params())


# TODO duplicate code
def show_dialog(x, y, width, header="", content=None):
    write_xy(x, y, colored("┌" + ("─" * width) + "┐", "white", "on_blue"))
    y += 1
    for idx, s in enumerate(header):
        write_xy(x, y, colored("│", "white", "on_blue"))
        write(colored(extent(s, width), "white", "on_blue"))
        write(colored("│", "white", "on_blue"))
        y += 1
    if content is not None:
        write_xy(x, y, colored("│", "white", "on_blue"))
        write(colored(extent(content, width), "white", "on_red", True))
        write(colored("│", "white", "on_blue"))
        y += 1
    write_xy(x, y, colored("└" + ("─" * width) + "┘", "white", "on_blue"))


# TODO make it reusable
def file_dialog(repo: Repository, conf: Config):
    width = min(cols() - 2, 40)
    y = rows() // 2 - 2
    x = cols() // 2 - width // 2
    s = conf.get_export_path(os.path.join(str(Path.home()), "papers.html"))
    while True:
        val = s + "▃"
        if len(s) >= width:
            val = val[-width:]
        show_dialog(x, y, width, header=["Filename for export:"], content=val)
        key = read_key()
        if key is None:
            continue
        if ord(key) == 27:
            break
        elif ord(key) == 10:
            conf.set_export_path(s)
            export_repository_html(repo, s)
            show_dialog(x, y, width, header=["Repository exported to:", s[:width]])
            read_key()
            break
        elif ord(key) == 127:
            s = s[:-1]
        elif key.isprintable():
            s += key


def run_ui(args, repo: Repository, conf: Config):
    papers = repo.list()
    if len(papers) == 0:
        print("Repository is empty.")
        sys.exit(0)

    if len(args) > 0:
        papers = filter_list(papers, args[0])

    # Retrieve latest version of papr from GitHub.
    latest_version(conf.set_latest_version)

    ui_main_or_search_loop(papers, repo, conf)
    cursor_on()
    print()





import sqlite3
from .v1_paper import Paper

SQLITE_FILE = "paper.db"
MAX_SUPPORT_DB_VERSION = 2


class Db:
    @staticmethod
    def create(path):
        conn = sqlite3.connect(path + "/" + SQLITE_FILE)
        c = conn.cursor()
        # SQL queries for version 1.
        c.execute("CREATE TABLE papers (idx integer primary key, json text)")
        # SQL queries for version 2.
        Db.version_2_commands(conn, c)
        conn.commit()
        conn.close()

    def get_connection(self):
        return sqlite3.connect(self.filename())

    @staticmethod
    def version_2_commands(conn, c):
        c.execute("CREATE TABLE config (version text)")
        data = (2,)
        c.execute("INSERT INTO config (version) values (?)", data)
        conn.commit()

    @staticmethod
    def version_3_commands(conn, c):
        pass

    @staticmethod
    def upgrade_from_1(conn, c):
        Db.version_2_commands(conn, c)

    @staticmethod
    def upgrade_from_2(conn, c):
        Db.version_3_commands(conn, c)

    def upgrade(self):
        # Version 1: does not has a table config.
        # Version 2: has a table config with a column config which has value 2.
        conn = self.get_connection()
        c = conn.cursor()
        try:
            c.execute("SELECT version FROM config")
            # Nothing to do as the database already has a table to store the config.
        except sqlite3.OperationalError:
            # Upgrade database to version 2.
            Db.upgrade_from_1(conn, c)
        # At this point we have at least version 2.
        conn.close()

    def __init__(self, path):
        self.path = path

    @staticmethod
    def get_version(c):
        return int(c.execute("SELECT version FROM config").fetchone()[0])

    def check_version(self):
        self.upgrade()
        conn = self.get_connection()
        c = conn.cursor()
        v = Db.get_version(c)
        conn.close()
        return v <= MAX_SUPPORT_DB_VERSION

    def filename(self):
        return self.path + "/" + SQLITE_FILE

    def list(self):
        conn = self.get_connection()
        c = conn.cursor()
        r = [Paper.from_json(j[0], j[1]) for j in sorted([i for i in c.execute("SELECT idx, json FROM papers")])]
        conn.close()
        return r

    def next_id(self):
        conn = self.get_connection()
        c = conn.cursor()
        r = [i[0] for i in c.execute("SELECT idx FROM papers")]
        conn.close()
        if len(r) == 0:
            return 1
        return max(r) + 1

    def add_paper(self, p: Paper):
        conn = self.get_connection()
        c = conn.cursor()
        data = (p.idx(), p.as_json())
        c.execute("INSERT INTO papers (idx, json) VALUES (?, ?)", data)
        conn.commit()
        conn.close()

    def get(self, idx):
        conn = self.get_connection()
        c = conn.cursor()
        r = c.execute("SELECT json FROM papers WHERE idx=" + str(idx))
        r = r.fetchone()
        conn.close()
        if not r:
            return None
        return Paper.from_json(idx, r[0])

    def update_paper(self, p: Paper):
        conn = self.get_connection()
        c = conn.cursor()
        data = (p.as_json(), p.idx())
        c.execute("UPDATE papers SET json = ? WHERE idx = ?", data)
        conn.commit()
        conn.close()




from .termout import print_header, print_papers
from .repository import Repository


def cmd_list(repo: Repository):
    r = repo.list()
    if len(r) == 0:
        print("empty")
    else:
        print_header()
        print_papers(r)





import sys

from .config import Config

VALID_KEYS = ["viewer", "browser", "browser_params"]


def error():
    print("Possible keys:")
    print(", ".join(VALID_KEYS))
    sys.exit(1)


def cmd_config(args, conf: Config):
    if len(args) < 2:
        print("You need to specify a key and a value.")
        error()

    key = args[0]
    if key not in VALID_KEYS:
        print("You specified an invalid config key.")
        error()

    val = args[1].strip()
    if key == "viewer":
        conf.set_viewer(val)
        print("Set viewer to value:", val)
    elif key == "browser":
        conf.set_browser(val)
        print("Set browser to value:", val)
    elif key == "browser_params":
        conf.set_browser_params(args[1:])
        print("Set browser parameters to:", args[1:])
    else:
        print("Unknown key.")
        error()






import sys

from .termout import rows


def write(s):
    sys.stdout.write(s)
    sys.stdout.flush()


def cursor_up(n):
    # http://tldp.org/HOWTO/Bash-Prompt-HOWTO/x361.html
    write('\033[' + str(n) + 'A')


def cursor_down(n):
    # http://tldp.org/HOWTO/Bash-Prompt-HOWTO/x361.html
    write('\033[' + str(n) + 'B')


#def cursor_right(n):
#    # http://tldp.org/HOWTO/Bash-Prompt-HOWTO/x361.html
#    write('\033[' + str(n) + 'C')


def cursor_bottom():
    write('\033[' + str(rows()) + 'B')
    write('\033[' + str(n) + 'B')


def cursor_top_left():
    write('\033[0;0H')


def cursor_off():
    write("\x1b[?25l")


def cursor_on():
    write("\x1b[?25h")


def cursor_goto(x, y):
    write('\033[' + str(y) + ";" + str(x) + 'H')


def write_xy(x, y, s):
    cursor_goto(x, y)
    write(s)



import re
import urllib
import urllib.request
from bs4 import BeautifulSoup


def parse_arxiv_abstract(data):
    h = BeautifulSoup(data, "html.parser")
    r = h.find_all("blockquote", {"class": "abstract"})
    if len(r) > 0:
        for i in r[0].find_all("span"):
            i.decompose()
        return r[0].get_text().replace("\n", " ").strip()
    return None


def parse_with_re(data, regx):
    d = data.decode("utf-8", errors="ignore").replace("\n", " ")
    m = regx.match(d)
    if m is not None:
        return m.group(1)
    return None


def parse_openreview_abstract(data):
    c = re.compile(r".*Abstract:</strong>\s*<span[^>]+>(.*?)</span>.*")
    return parse_with_re(data, c)


def parse_neurips_abstract(data):
    c = re.compile(r".*<h4>Abstract</h4>\s*<p>(.*?)</p>.*")
    return parse_with_re(data, c)


def parse_abstract(data):
    r = parse_arxiv_abstract(data)
    if r is None:
        r = parse_openreview_abstract(data)
    if r is None:
        r = parse_neurips_abstract(data)
    return r


def load_abstract(url):
    # TODO merge this with cmd_fetch.py
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36"
    }
    req = urllib.request.Request(url, headers=headers)
    rsp = urllib.request.urlopen(req)
    data = rsp.read()
    data = parse_abstract(data)
    return data


if __name__ == "__main__":
    print("arxiv")
    print(load_abstract("https://arxiv.org/abs/1905.05454"))
    print("openreview")
    print(load_abstract("https://openreview.net/forum?id=HkNDsiC9KQ"))



import os
import termios
import sys
from select import select

from .termout import cols, rows


def read_key():
    n_cols = cols()
    n_rows = rows()

    fd = sys.stdin.fileno()
    oldterm = termios.tcgetattr(fd)
    newattr = termios.tcgetattr(fd)
    newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSANOW, newattr)
    c = []
    try:
        while True:
            r1, r2, r3 = select([fd], [], [], 0.05)
            if not r1 and not r2 and not r3: # timeout -> check if size of terminal has changed
                if n_cols != cols() or n_rows != rows():
                    c = [0, 1, 2]
                    break
            else:
                c = [i for i in os.read(fd, 50)]
                break
    except IOError:
        pass
    termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
    if len(c) == 1:
        return chr(c[0])
    else:
        if c == [0, 1, 2]:  # redraw screen
            return '~'
        elif c == [27, 91, 65]:
            return '#'
        elif c == [27, 91, 66]:
            return '+'
        elif c == [27, 91, 49, 59, 50, 65]: # shift + up
            return '['
        elif c == [27, 91, 49, 59, 50, 66]: # shift + down
            return ']'
        elif c == [27, 91, 54, 126]: # page down
            return '§'
        elif c == [27, 91, 53, 126]: # page up
            return '$'
        #print(c)
        #while True:
        #    pass




import os
import re
import time
from pathlib import Path
from subprocess import Popen, DEVNULL

from .paper import Paper
from .edit import create_tmp_file
from .html import export_papers_html


def re_exists_in(p: Paper, q):
    try:
        r = re.compile(q.lower())
        if r.search(p.title().lower()):
            return True
        return False
    except Exception:
        return False


def substrings_exist_in(paper: Paper, query):
    for q in query.split():
        if paper.title().lower().find(q.lower()) < 0:
            return False
    return True


def merge_overlaps(positions):
    if len(positions) < 2:
        return positions
    p = positions[:]
    p.sort()
    p_beg, p_end = p[0]
    r_beg, r_end = p[1]
    if r_beg < p_end:
        return merge_overlaps(p[2:] + [(p_beg, max(p_end, r_end))])
    else:
        return [p[0]] + merge_overlaps(p[1:])


def highlight_query(paper: Paper, query):
    positions = []
    for q in query.split():
        start_pos = 0
        while True:
            pos = paper.title().lower().find(q.lower(), start_pos)
            if pos == -1:
                break
            positions.append((pos, pos + len(q)))
            start_pos = pos + len(q)
    return merge_overlaps(positions)


def filter_list(papers, query):
    return [p for p in papers if substrings_exist_in(p, query) or len(query) == 0]


def filter_list_re(papers, query):
    return [p for p in papers if re_exists_in(p, query) or len(query) == 0]


def show_pdf(p: Paper, repo_path, path_viewer):
    try:
        Popen([path_viewer, repo_path + "/" + p.filename()], stderr=DEVNULL, stdout=DEVNULL)
    except:
        pass


def show_in_browser(repo, p: Paper, browser, browser_params=[]):
    tmp = create_tmp_file(ext="html")
    export_papers_html(repo, [p], tmp, with_pdf_link=True)
    try:
        # TODO configure browser
        Popen([browser] + browser_params + ["file://" + tmp], stderr=DEVNULL, stdout=DEVNULL)
    except:
        pass




import sys


def cmd_init(repo) -> None:
    # Check that the current directory isn't already a repository.
    if repo.is_local_repository():
        print("You are already in a repository.", file=sys.stderr)
        sys.exit(1)
    # Create repository.
    repo.init()
    print("Repository created.")





import sys
from .repository import Repository
from .paper import Paper


def cmd_add(args, repo: Repository):
    if len(args) == 0:
        print("You need to specify a filename.")
        sys.exit(1)
    filename = args[0]
    idx = repo.next_id()
    if len(args) > 1:
        idx = int(args[1])
    title = filename
    if title.rfind(".") >= 0:
        title = title[:title.rfind(".")]
    p = Paper(idx=idx, filename=filename, title=title)
    repo.add_paper(p)
    print("Added paper.")
    print("Filename:", filename)



import os
import sys

from .v1_db import Db
from papr.lib.config import Config

REPO_META = ".paper"


class RepositoryV1:
    def __init__(self, config: Config):
        self.db = None
        self.config = config

        # determine base_path and path for REPO_META
        self._determine_paths()

        if self.repo_meta_path is None:
            return
        self.db = Db(self.repo_meta_path)
        if not self.db.check_version():
            print("Found incompatible version.")
            sys.exit(1)

    def pdf_path(self):
        return self.repo_pdf_path

    def _determine_paths(self):
        # First check if the current working directory is a repository.
        self.repo_pdf_path = os.getcwd()  # $PWD
        self.repo_meta_path = self.repo_pdf_path + "/" + REPO_META  # $PWD/.paper/
        self.local = True

        if not os.path.exists(self.repo_meta_path):
            # If it does not exist read the location of the repository from
            # the configuration.
            self.repo_pdf_path = self.config.get("default_repo")
            self.repo_meta_path = self.repo_pdf_path + "/" + REPO_META
            self.local = False

        if not os.path.exists(self.repo_meta_path):
            self.repo_pdf_path = None
            self.repo_meta_path = None
            self.local = None

    def list(self):
        """
        Returns a list of all papers.
        :return:
        """
        return self.db.list()



import urllib.request
import re
from threading import Thread


def latest_version(callback):
    t = Thread(target=version_getter, args=(callback,))
    t.start()


def version_getter(callback):
    try:
        URL = "https://raw.githubusercontent.com/daniel-e/papr/master/setup.py"
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36"
        }
        req = urllib.request.Request(URL, headers=headers)
        rsp = urllib.request.urlopen(req)
        if rsp.status / 100 == 2:
            data = rsp.read().decode("utf-8", errors="ignore")
            c = re.compile(r"^\s*version=\"([0-9]+\.[0-9]+\.[0-9]+)\",$")
            for i in data.split("\n"):
                m = c.match(i)
                if m is not None:
                    callback(m.group(1))
                    break
    except:
        pass


def __callback(s):
    print(s)


if __name__ == "__main__":
    latest_version(__callback)



# The last element is selected at the beginning.
class ScrollView:
    def __init__(self, n_elements, rows, selected):
        self.n_elemens = n_elements
        self._rows = rows
        self._selected = selected    # selected is in [0..n_elements-1]
        self.first_element_in_view = max(0, n_elements - rows)

    def pagedown(self):

        if self.first_element_in_view + self._rows < self.n_elemens:
            self.first_element_in_view += self._rows
            self._selected += self._rows
        else:
            self._selected = self.n_elemens - 1
        pass

    def pageup(self):
        if self.first_element_in_view == 0:
            self._selected = 0
        elif self.first_element_in_view - self._rows > 0:
            self.first_element_in_view -= self._rows
            self._selected -= self._rows
        else:
            n = self.first_element_in_view
            self.first_element_in_view = 0
            self._selected -= n
        pass

    def up(self, n=1):
        while n > 0:
            if self.first_element_in_view == self._selected:
                self.first_element_in_view = max(self.first_element_in_view - 1, 0)
            self._selected = max(0, self._selected - 1)
            n -= 1

    def down(self, n=1):
        while n > 0:
            if self._selected >= self.first_element_in_view + self._rows - 1:
                self.first_element_in_view = min(self.first_element_in_view + 1, self.n_elemens - self._rows)
            self._selected = min(self._selected + 1, self.n_elemens - 1)
            n -= 1

    def selected(self):
        return self._selected

    def first(self):
        return self.first_element_in_view

    def end(self):
        return self.first_element_in_view + self._rows

    def rows(self):
        return self._rows


import os
import sys

import termcolor


def write(s):
    sys.stdout.write(s)
    sys.stdout.flush()


def _hr():
    cols, _ = os.get_terminal_size(0)
    print("─" * cols)


def rows():
    _, rows = os.get_terminal_size(0)
    return rows


def cols():
    cols, _ = os.get_terminal_size(0)
    return cols


def empty_line():
    cols, _ = os.get_terminal_size(0)
    return " " * cols


# TODO duplicate
def print_header():
    cols, _ = os.get_terminal_size(0)
    print("Id    Title " + " " * (cols - 12 - 8) + "Stars" + "  ")
    _hr()
    return cols


def print_paper(paper, selected=False):
    print(paper_line(paper, selected))


def paper_line(paper, selected=False):
    cols, _ = os.get_terminal_size(0)
    # 5 columns for id
    # 1 columns for space
    # 2 columns at the end of the line for notes
    # n columns for title

    stars = ""
    if paper.stars() >= 0:
        stars = " " * (5 - paper.stars()) + "*" * paper.stars()
    len_stars = len(stars) + 3

    tags = ""
    len_tags = 0
    if len(paper.tags()) > 0:
        tags = ",".join(paper.tags())
        len_tags = len(tags) + 3

    n = cols - (5 + 1 + 2) - len_stars - len_tags
    f = paper.title()
    if len(f) > n:
        f = f[:n - 3] + "..."

    orig_f = f

    h = paper.highlights()[:]
    h.sort(reverse=True)
    for beg, end in h:
        if beg < len(f):
            f = f[:beg] + colored("highlight", True, f[beg:end]) + colored("line", selected, f[end:])

    m = "{:5d} ".format(paper.idx())
    s = colored("line", selected, m + f)
    if len(m) + len(orig_f) < cols:
        s = s + colored("line", selected, " " * (cols - len(m + orig_f) - 2 - len_stars - len_tags))

    if len_tags > 0:
        s = s + colored("line", selected, " [") + colored("tags", selected, tags) + colored("line", selected, "]")
    if len_stars > 0:
        s = s + colored("line", selected, " [" + stars + "]")

    if paper.has_notes():
        s += colored("line", selected, " ✺")
    else:
        s += colored("line", selected, "  ")

    return s


def colored(typ, selected, s):
    mapping = {"line": "white", "tags": "green"}
    if typ == "highlight":
        return termcolor.colored(s, "yellow", 'on_yellow', attrs=["bold"])

    if selected:
        return termcolor.colored(s, mapping[typ], 'on_red', attrs=["bold"])
    else:
        return termcolor.colored(s, mapping[typ])


def print_papers(papers):
    for paper in papers:
        print_paper(paper)




import math
import sys
import urllib
import urllib.request
import os
import shutil
import re
import argparse

import yaml
from bs4 import BeautifulSoup

from .html import export_repository_html
from .paper import Paper
from .edit import editor
from .abstract import parse_abstract


NEWTAG="unread"


def title_as_filename(s):
    # lower case
    s = re.sub(r'\s+', "_", s.lower())
    # remove everything except a-z0-9
    s = re.sub(r'[^a-z0-9]', "_", s)
    # remove double _
    s = re.sub(r'_+', '_', s)
    # remove trailing _
    s = re.sub(r'_+$', '', s)
    return s


def normalize_title(s):
    return re.sub(r'\s+', " ", s.replace("\n", " "))


def prepare_data(title: str, repo):
    title = normalize_title(title)
    idx = repo.next_id()
    filename = "{:05d}_{}.pdf".format(idx, title_as_filename(title))
    abspath = repo.pdf_path() + "/" + filename
    return title, idx, filename, abspath


def add_local_file(f: str, title, repo, tags):
    tmp_title = ""
    if len(title) == 0:
        tmp_title = determine_title_via_vim()
    else:
        tmp_title = title

    title, idx, filename, abspath = prepare_data(tmp_title, repo)
    if not os.path.exists(abspath):
        shutil.copy(f, abspath)
    p = Paper(idx=idx, filename=filename, title=title, tags=[NEWTAG]+tags)
    repo.add_paper(p)
    print("Added paper.")
    print("Title   :", title)
    print("Filename:", filename)


def is_text_or_html(s):
    return s is not None and (s.lower().find("text") >= 0 or s.lower().find("html") >= 0)


def white():
    return "\x1b[97;1m"


def reset():
    return "\x1b[0m"


def bar(p):
    chars = "▏▎▍▌▋▊▉█"
    N = 40
    l = p * N / 100.0
    todo = "." * (N - int(l))
    done = "█" * int(l)

    f = l - math.floor(l)
    if f > 0:
        todo = "." * (N - int(l) - 1)
        done = done + chars[int(f * len(chars))]

    r = str(int(255 - 255 * p / 100.0))
    g = str(int(255 * p / 100.0))
    b = "0"
    sys.stdout.write("\r|\x1b[38;2;{};{};{}m{}\x1b[0m{}| {:3d}% ".format(r, g, b, done, todo, int(p)))
    sys.stdout.flush()


def download(rsp, no_progress_bar=False):
    n = rsp.getheader("content-length")
    data = bytes()
    if n is not None:
        n = int(n)
        s = n
        while n > 0:
            t = rsp.read(min(1024, n))
            data += t
            n -= min(1024, n)
            if not no_progress_bar:
                bar(100.0 * (s - n) / s)
        if not no_progress_bar:
            print()
    else:
        chars = "|/-|\\"
        n = 0
        while True:
            t = rsp.read(10240)
            if len(t) == 0:
                break
            data += t
            if not no_progress_bar:
                sys.stdout.write("\rloading " + chars[n % len(chars)])
                sys.stdout.flush()
            n += 1
        if not no_progress_bar:
            print("\r          ")
    return data


def parse_page(data):
    h = BeautifulSoup(data, "html.parser")
    r = h.find_all("meta")
    title = [i.get("content") for i in r if i.get("name") == "citation_title"]
    pdfurl = [i.get("content") for i in r if i.get("name") == "citation_pdf_url"]
    abstract = parse_abstract(data)
    if len(title) > 0 and len(pdfurl) > 0:
        return title[0], pdfurl[0], abstract
    return None, None, None


def determine_title_via_vim():
    msg = "# Please enter the title of the paper.\n\n"
    msg = editor(msg, 2)
    s = ""
    for line in msg.split("\n"):
        c = re.compile(r"\s*#.*")
        m = c.match(line)
        if m is None:
            s = (s + " " + line.strip()).strip()
    if len(s) == 0:
        s = "(emtpy)"
    return s


def do_fetch_download(repo, url, title, tags):
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36"
    }
    req = urllib.request.Request(url, headers=headers)
    rsp = urllib.request.urlopen(req)
    typ = rsp.getheader("content-type")
    if not is_text_or_html(typ) and title == "":  # for PDF files we need a title
        title = determine_title_via_vim()
    data = download(rsp, is_text_or_html(typ))
    abstract = ""
    if is_text_or_html(typ):
        t, pdfurl, abstract = parse_page(data)
        if t is None:
            print("Error")
            sys.exit(1)
        title = t
        req = urllib.request.Request(pdfurl)
        rsp = urllib.request.urlopen(req)
        data = download(rsp)

    title, idx, filename, abspath = prepare_data(title, repo)

    k = open(abspath, "w")
    k.buffer.write(data)
    k.close()

    p = Paper(idx=idx, filename=filename, title=title, tags=[NEWTAG]+tags)
    p.set_url(url)
    p.set_abstract(abstract)
    repo.add_paper(p)
    print("{}{}{} {}".format(white(), "   Title:", reset(), title))
    print("{}{}{} {}".format(white(), "Filename:", reset(), filename))
    print("{}{}{} {}".format(white(), "Abstract:", reset(), abstract))
    print("{}{}{}".format(white(), "Added paper.", reset()))
    print()


def do_fetch(location, title, repo, tags):
    if title is None:
        title = ""
    if os.path.exists(location):
        add_local_file(location, title, repo, tags=tags)
    else:
        do_fetch_download(repo, url=location, title=title, tags=tags)


def cmd_fetch(args, repo):
    parser = argparse.ArgumentParser(description="Fetch a paper.")
    parser.add_argument("--urls", nargs=1, required=False, metavar="filename", help="Load papers from URLs stored in a file. (one URL per line)")
    parser.add_argument("-t", nargs=1, required=False, type=str, help="Title of paper.")
    parser.add_argument("--tags", nargs=1, required=False, type=str, help="Comma separated list of tags.")
    parser.add_argument("location", nargs='*', type=str)
    #parser.print_help()
    args = parser.parse_args(args)

    title = ""
    if args.t is not None:
        title = args.t[0]

    tags = []
    if args.tags is not None:
        tags = [i.strip() for i in args.tags[0].split(",")]

    if args.urls is not None:
        with open(args.urls[0], "r") as f:
            for line in f:
                s = line.strip()
                if len(s) > 0:
                    print("fetching", s, "...", file=sys.stderr)
                    do_fetch(s, title, repo, tags)
    else:
        for location in args.location:
            try:
                do_fetch(location, title, repo, tags)
            except:
                pass


def cmd_export_html(args, repo):
    parser = argparse.ArgumentParser(description="Export the database into an HTML file.")
    parser.add_argument("--with-note", action='store_true', help="Export papers only if they contain notes.")
    parser.add_argument("--with-summary", action='store_true', help="Export papers only if they contain a summary.")
    parser.add_argument("-n", type=int, default=-1, help="Export the n most recent papers.")
    parser.add_argument("filename", nargs=1, type=str)
    args = parser.parse_args(args)
    export_repository_html(repo, args.filename[0], with_notes=args.with_note, with_summary=args.with_summary, n=args.n)


def cmd_export_yml(args, repo):
    parser = argparse.ArgumentParser(description="Export the database into a YML file.")
    parser.add_argument("filename", nargs=1, type=str, description="YML file.")
    args = parser.parse_args(args)

    # TODO export summary (paper.summary()) and notes (paper.msg()) as well

    path = args.filename[0]
    all_meta = []
    for paper in repo.list():
        meta = {
            "title": paper.title(),
            "stars": paper.stars(),
            "idx": paper.idx(),
            "pdf": paper.filename(),
            "tags": ",".join(paper.tags()),
            "abstract": paper.abstract(),
            "url": paper.url(),
        }
        all_meta.append(meta)
    with open(path, "wt") as f:
        yaml.dump(all_meta, f)



import tempfile
import os
import unittest
from subprocess import call

import termcolor

from .paper import Paper


def create_tmp_file(content="", ext="tmp.md"):
    _, pth = tempfile.mkstemp("." + ext)
    f = open(pth, "wt")
    f.write(content)
    f.close()
    return pth


# Open the string stored in _msg_ in an editor.
# Returns: content of edited file
def editor(msg, n=None):
    e = os.getenv('EDITOR', 'vim')
    pth = create_tmp_file(msg)
    cmd = [e]
    if n is not None and e == "vim":
        cmd.append("+" + str(n))
    cmd.append(pth)
    call(cmd)
    f = open(pth, "r")
    msg = f.read()
    f.close()
    return msg


def less(msg, args=[]):
    pth = create_tmp_file(msg)
    cmd = ["less", "-c"] + args + [pth]
    call(cmd)


def _edit_file(filename):
    call([os.getenv('MARKUP_EDITOR', 'vim'), filename])


def edit_notes_of_paper(repo, p: Paper):
    repo.path_for_paper(p).mkdir(parents=True, exist_ok=True)
    filename = repo.notes_filename(p)
    _edit_file(filename)
    p.set_has_notes(len(filename.read_text().strip()) > 0)
    repo.update_paper(p)


def edit_summary_of_paper(repo, p: Paper):
    repo.path_for_paper(p).mkdir(parents=True, exist_ok=True)
    filename = repo.summary_filename(p)
    _edit_file(filename)
    p.set_has_summary(len(filename.read_text().strip()) > 0)
    repo.update_paper(p)


def tags_of_paper(repo, p: Paper):
    msg = "COMMA SEPARATED LIST OF WORDS\n" + ",".join(p.tags())
    msg = editor(msg, 2)
    pos = msg.find("\n")
    if pos >= 0:
        msg = msg[pos+1:]
    msg = msg.replace("\n", ",")
    tags = [j for j in [i.strip().lower() for i in msg.split(",")] if len(j) > 0]
    p.set_tags(tags)
    repo.update_paper(p)


def abstract_of_paper(p: Paper):
    abstract = p.abstract()
    if abstract == "":
        abstract = "No abstract available."
    less(abstract)


def details_str(key, val):
    return key + ":\n" + ("=" * (len(key) + 1)) + "\n" + str(val) + "\n\n"


def details_of_paper(p: Paper):
    d = p.as_nice_dict()
    t = ""
    for key, val in {i: j for i, j in d.items() if i != "Notes"}.items():
        t += details_str(key, val)
    t += details_str("Notes", d.get("Notes", ""))
    less(t)


def bar(n, maxn, maxlen=30):
    return "█" * (maxlen * n // maxn)


def list_of_tags(repo):
    c = repo.all_tags()
    maxtaglen = max([len(tag) for tag, _ in c])
    maxn = max([n for _, n in c])
    msg = ""
    for tag, n in c:
        msg += tag + (" " * (maxtaglen - len(tag))) + " | " + "{:4}".format(n) + " " + bar(n, maxn) + "\n"
    msg += "\n\nPress q to quit."
    less(msg)


def edit_title(p: Paper, r):
    msg = editor(p.title()).strip()
    if msg != p.title():
        p.set_title(msg)
        r.update_paper(p)


def wrap_lines(s, width):
    lines = []
    while len(s) > width:
        pos = width
        while s[pos] != ' ':
            pos -= 1
        lines.append(s[:pos])
        s = s[pos+1:]
    if len(s) > 0:
        lines.append(s)
    return "\n".join(lines)


def show_summaries(repo, width):
    papers = repo.list()
    msg = ""
    width -= 1  # less has trouble to correctly show the content when not doing this
    for p in papers:
        if len(p.summary().strip()) > 0:
            t = p.title()
            d = max(0, width - len(t))
            msg += termcolor.colored(t + " " * d, "white", "on_blue", attrs=["bold"]) + "\n"
            msg += "─" * width + "\n"
            msg += wrap_lines(p.summary(), width) + "\n\n"
    msg += termcolor.colored("\nPress q to quit.", "white", "on_red", attrs=["bold"])
    less(msg, ["-r"])


class TestEdit(unittest.TestCase):
    def test_insert_and_remove_markdown(self):
        msg = ""
        msg = insert_markdown_help(msg)
        self.assertEqual(msg, "\n" + MARKDOWN_HELP)
        msg = remove_markdown_help(msg)
        self.assertEqual(msg, "")

        msg = "A"
        msg = insert_markdown_help(msg)
        self.assertEqual(msg, "A\n" + MARKDOWN_HELP)
        msg = remove_markdown_help(msg)
        self.assertEqual(msg, "A")

        msg = "A\n"
        msg = insert_markdown_help(msg)
        self.assertEqual(msg, "A\n\n" + MARKDOWN_HELP)
        msg = remove_markdown_help(msg)
        self.assertEqual(msg, "A\n")


if __name__ == "__main__":
    unittest.main()


import os

import markdown
from .paper import Paper


def map_header(header):
    color = {
        "Abstract": "alert-primary",
        "Notes": "alert-success",
        "Summary": "alert-danger"
    }
    return color.get(header, "btn-outline-dark")


def add_header(header):
    if header is not None:
        return '<div class="alert ' + \
               map_header(header) + ' btn-sm" role="alert" style="text-align: center; font-weight:bold">' + header + \
               '</div>'
    return ""


def ashtml(s: str, header=None):
    return "" if len(s.strip()) == 0 else \
        "<tr><td>{}{}</td></tr>".format(add_header(header), markdown.markdown(s, extensions=['tables']))


def stars(n):
    return ("★"*n) + ("☆"*(5-n))


def export_paper_html(repo, p: Paper, f, with_pdf_link=False):
    pdf_link = ""
    if with_pdf_link:
        pdf_link = \
            "<a href='file://{}' class='btn btn-outline-secondary btn-sm role='button'>local pdf</a>" \
            .format(os.path.join(repo.pdf_path(), p.filename()))
    extern_link = "extern link"
    if p.url().find("arxiv.org") != -1:
        extern_link = "on arXiv"
    print(
        "<thead class='thead-dark'><tr><th>{}</th></tr></thead>"  # title
        "<tr>"
        "  <td>{}&nbsp;&nbsp;&nbsp;<a href='{}' class='btn btn-outline-secondary btn-sm role='button'>{}</a> {}</td>"
        "</tr>"  # url
        "{}"  # abstract
        "{}"  # summary
        "{}"  # notes
            .format(p.title(), stars(p.stars()), p.url(), extern_link, pdf_link,
                    ashtml(p.abstract(), "Abstract"),
                    ashtml(p.summary(), "Summary"),
                    ashtml(p.msg(), "Notes")
                    ), file=f)


def export_repository_html(repo, dstfile, with_notes=False, with_summary=False, n=-1):
    papers = reversed(repo.list())
    export_papers_html(repo, papers, dstfile, with_notes, with_summary, n)


def export_papers_html(repo, papers, dstfile, with_notes=False, with_summary=False, n=-1, with_pdf_link=False):
    mydir = os.path.dirname(__file__)
    header = open(os.path.join(mydir, "../html/header.html"), "r").read()
    footer = open(os.path.join(mydir, "../html/footer.html"), "r").read()
    with open(dstfile, "w") as f:
        print(header, file=f)
        print('<table class="table">', file=f)
        cnt = 0
        for p in papers:
            include = not(with_notes or with_summary)
            if with_summary:
                include = include or p.has_summary()
            if with_notes:
                include = include or p.has_notes()
            if include:
                cnt += 1
                if n != -1 and cnt > n:
                    break
                export_paper_html(repo, p, f, with_pdf_link=with_pdf_link)
        print('</table>', file=f)
        print(footer, file=f)



import json


class Paper:
    def __init__(self, idx, filename, title, tags=[]):
        self._idx = idx
        self._filename = filename
        self._title = title
        self._msg = ""
        self._tags = tags
        self._stars = 0
        self._url = ""
        self._abstract = ""
        self._hidden = False
        self._summary = ""
        self._highlights = []  # for search

    @staticmethod
    def from_json(idx, jsonstr):
        p = Paper(idx, filename=None, title=None)
        p._idx = idx
        data = json.loads(jsonstr)
        p._filename = data["filename"]
        p._title = data["title"]
        p._msg = data.get("msg", "")          # notes -> only relevant for v1 repositories
        p._summary = data.get("summary", "")  # only relevant for v1 repositories
        p._tags = [i for i in data.get("tags", "").split(",") if len(i) > 0]
        p._stars = int(data.get("stars", "0"))
        p._abstract = data.get("abstract", "")
        p._url = data.get("url", "")
        p._hidden = False if data.get("hidden", "n") == "n" else True
        return p

    def as_nice_dict(self):
        mapping = {
            "filename": "Filename",
            "title": "Title",
            "msg": "Notes",
            "summary": "Summary",
            "tags": "Tags",
            "stars": "Voting",
            "abstract": "Abstract",
            "url": "URL"
        }
        r = {mapping.get(i, "<unknown>"): j for i, j in self._as_dict().items()}
        return r

    def _as_dict(self):
        d = {
            "filename": self._filename,
            "title": self._title,
            "msg": "",  # with repo v2 not stored in sqlite3 db anymore
            "summary": "",  # with repo v2 not stored in sqlite3 db anymore
            "tags": ",".join(self._tags),
            "stars": self._stars,
            "abstract": self._abstract,
            "url": self._url,
            "hidden": "y" if self._hidden else "n"
        }
        return d

    def as_json(self):
        return json.dumps(self._as_dict())

    def remove_tag(self, tag):
        self._tags = [i for i in self._tags if i != tag]

    def idx(self):
        return self._idx

    def filename(self):
        return self._filename

    def title(self):
        return self._title

    def set_title(self, t):
        self._title = t

    def msg(self):  # Notes
        return self._msg

    def update_msg(self, msg):
        self._msg = msg

    def summary(self):
        return self._summary

    def update_summary(self, summary):
        self._summary = summary

    def has_notes(self):
        return len(self._msg) > 0

    def has_summary(self):
        return len(self.summary()) > 0

    def tags(self):
        """
        Returns a list of tags.
        :return:
        """
        return sorted(self._tags)

    def set_tags(self, tags):
        self._tags = tags

    def stars(self):
        return self._stars

    def set_stars(self, n):
        self._stars = max(0, min(5, n))

    def set_abstract(self, abstract):
        self._abstract = ""
        if abstract is not None:
            self._abstract = abstract

    def set_url(self, url):
        self._url = url

    def abstract(self):
        return self._abstract

    def url(self):
        return self._url

    def set_highlights(self, positions):
        self._highlights = positions

    def highlights(self):
        return self._highlights

    def hidden(self):
        return self._hidden

    def hide(self, value=True):
        self._hidden = value



from .repository import Repository
from .tools import show_pdf


def cmd_last(args, repo: Repository):
    p = -1
    if len(args) > 0:
        p = -int(args[0])
    r = repo.list()
    if len(r) == 0:
        print("empty")
        return
    show_pdf(r[p], repo.pdf_path())





import sys

from .repository import Repository
from .termout import print_papers
from .tools import filter_list


def cmd_search(args, repo: Repository):
    if len(args) == 0:
        print("You need to specify a query.")
        sys.exit(1)
    r = repo.list()
    if len(r) == 0:
        print("empty")
        return
    query = args[0]
    r = filter_list(r, query)
    if len(r) == 0:
        print("Not found.")
    else:
        print_papers(r)





import os
import pathlib
import sys
from collections import Counter

import yaml

from .db import Db
from .config import Config
from .cmd_fetch import title_as_filename

DATA_PATH = "data"      # new in v2 repository
META_FILE = "meta.yml"  # new in v2 repository
REPO_META = ".paper"    # from v1 repository

# FILES
# ------------------------------------
# <root_path>/<REPO_META>/<META_FILE> -> has the version of the repository
# <root_path>/<DATA_PATH>/p00000_...  -> directory with notes.md and summary.md


def _limit_len(s, maxlen):
    while len(s) > maxlen:
        p = s.rfind("_")
        if p >= 0:
            s = s[:p]
        else:
            s = s[:maxlen]
    return s


def path_for_paper_data(paper, data_path, maxlen=96):
    dir_name = f"{paper.idx():05d}_{title_as_filename(paper.title())}"
    return data_path.joinpath(_limit_len(dir_name, maxlen))


class Repository:
    def __init__(self, config: Config):
        self.db = None
        self.config = config
        self._valid = False

        root_path = self._determine_root()
        if not self.is_valid():
            return

        self._repo_pdf_path = root_path
        self._repo_data_path = pathlib.Path(root_path).joinpath(DATA_PATH)
        self._repo_meta_path = pathlib.Path(root_path).joinpath(REPO_META)

        self.db = Db(self._repo_meta_path, self)
        if not self.db.check_version():
            print("Found incompatible version.")
            sys.exit(1)

    def is_local_repository(self):
        root_path = os.getcwd()
        return os.path.exists(root_path + "/" + REPO_META)  # $PWD/.paper/

    def pdf_path(self):
        return self._repo_pdf_path

    def is_v2_repo(self):
        # TODO check version in that file
        return self._repo_meta_path.joinpath(META_FILE).exists()

    def _determine_root(self):
        # First check if the current working directory is a repository.
        root_path = os.getcwd()
        if os.path.exists(root_path + "/" + REPO_META):  # $PWD/.paper/
            self._valid = True
            return root_path
        # If it does not exist read the location of the repository from
        # the configuration.
        root_path = self.config.get("default_repo")
        if os.path.exists(root_path + "/" + REPO_META):  # <config_repo_root>/.paper/
            self._valid = True
            return root_path

    def is_valid(self):
        return self._valid

    def init(self):
        """
        Creates directories, databases and updates configuration but does not change
        this isntance into a valid repository.
        :return:
        """
        pdf_path = os.getcwd()
        data_path = pathlib.Path(pdf_path).joinpath(DATA_PATH)
        meta_path = pathlib.Path(pdf_path).joinpath(REPO_META)
        meta_file = meta_path.joinpath(META_FILE)

        # Update configuration: set default repository.
        self.config.set_default_repo(pdf_path)
        # Create database.
        data_path.mkdir(parents=True, exist_ok=True)
        meta_path.mkdir(parents=True, exist_ok=True)

        data = {
            "repo_version": "2"
        }
        with open(meta_file, "wt") as f:
            yaml.dump(data, f)

        self.db = Db.create(meta_path)

    def path_for_paper(self, paper):
        return path_for_paper_data(paper, self._repo_data_path)

    def summary_filename(self, paper):
        return self.path_for_paper(paper).joinpath("summary.md")

    def notes_filename(self, paper):
        return self.path_for_paper(paper).joinpath("notes.md")

    def load_paper_data(self, paper):
        summary = ""
        notes = ""
        summary_file = self.summary_filename(paper)
        if summary_file.exists():
            summary = summary_file.read_text()
        notes_file = self.notes_filename(paper)
        if notes_file.exists():
            notes = notes_file.read_text()
        data = {
            "summary": summary,
            "notes": notes
        }
        return data

    def save_summary(self, paper, summary_str):
        self.path_for_paper(paper).mkdir(parents=True, exist_ok=True)
        self.path_for_paper(paper).joinpath("summary.md").write_text(summary_str)

    def save_notes(self, paper, notes_str):
        self.path_for_paper(paper).mkdir(parents=True, exist_ok=True)
        self.path_for_paper(paper).joinpath("notes.md").write_text(notes_str)

    def list(self):
        """
        Returns a list of all papers.
        :return:
        """
        return self.db.list()

    def next_id(self):
        return self.db.next_id()

    def add_paper(self, p):
        return self.db.add_paper(p)

    def get_paper(self, idx):
        return self.db.get(idx)

    def update_paper(self, p):
        self.db.update_paper(p)

    def all_tags(self, sorted_by_usage=True):
        # Get a list of list of tags.
        l = [paper.tags() for paper in self.list_partial()]
        # Flatten the list and count the occurrences of each tag.
        c = Counter([j for i in l for j in i])
        return c.most_common()
